;;;
;;; fun.scm --- getting fun with groups
;;;

(define (change-ith list i value)
  (if (not (pair? list))
      '()
      (if (= i 0)
          (cons value (cdr list))
          (cons (car list) (change-ith (cdr list) (- i 1) value)))))

; constructs product group g1 cross g2, including the new operation
(define (product-group . groups)
  (define (multiply-groups g1 g2)
    (let ((op1 (group-operation g1)) (op2 (group-operation g2))
          (g1 (group-elements g1)) (g2 (group-elements g2)))
      (let loop1 ((result '()) (g1 g1))
        (let loop2 ((result result) (g2 g2))
          (if (pair? g1)
              (if (pair? g2)
                  (loop2 (cons (list (car g1) (car g2)) result) (cdr g2))
                  (loop1 result (cdr g1)))
              (make-group result (lambda (x y) (list (op1 (car x) (car y)) (op2 (cadr x) (cadr y))))))))))

  (cond ((and (pair? groups) (null? (cdr groups)))
         (car groups))
        ((pair? groups)
         (multiply-groups (car groups) (apply product-group (cdr groups))))))

(define (line-to-cycle perm)
  (define (get-cycles perm unused)
    (let loop ((i 0))
      (cond ((= i (length unused))
             '())
            ((list-ref unused i)
             (let ((start i))
               (let loop2 ((result (list i)) (j start))
                 (set! unused (change-ith unused j #f))
                 (if (= (list-ref perm j) start)
                     (cons result (get-cycles perm unused))
                     (loop2 (cons (list-ref perm j) result) (list-ref perm j))))))
            (else
             (loop (+ i 1))))))

  (map reverse (get-cycles perm (make-list (length perm) #t))))

#|
;; this for some reason isn't tail recursive.
; list of generators
(define (generate-group generators operation)
  (define (make-hash-map element generators hash-table)
    (let loop ((to-try generators) (hash-table hash-table))
      (if (pair? to-try)
          (let* ((new (operation (car to-try) element))
                   (found (hash-table/get hash-table new #f)))
            (if (not found)
                (begin
                  (hash-table/put! hash-table new #t)
                  (loop (cdr to-try) (make-hash-map new generators hash-table))))
            (loop (cdr to-try) hash-table))))
    
    hash-table)
  
  (make-group
   (hash-table/key-list
      (fold-right (lambda (elt hash-table) (make-hash-map elt generators hash-table)) (make-equal-hash-table) generators))
   
   operation))
|#

(define (generate-group generators operation)
  (let ((elements (make-equal-hash-table))
        (to-see (make-queue)))
    (for-each (lambda (g)
                (hash-table/put! elements g #t)
                (enqueue!/unsafe to-see g))
              generators)
    (let lp ((e (dequeue!/unsafe to-see)))
      (for-each (lambda (a)
                  (let ((e1 (operation a e))
                        (e2 (operation e a)))
                    (if (not (hash-table/get elements e1 #f))
                        (begin
                          (hash-table/put! elements e1 #t)
                          (enqueue!/unsafe to-see e1)))
                    (if (not (hash-table/get elements e2 #f))
                        (begin
                          (hash-table/put! elements e2 #t)
                          (enqueue!/unsafe to-see e2)))))
                (hash-table/key-list elements))
      (if (queue-empty? to-see)
          (make-group (hash-table/key-list elements) operation)
          (lp (dequeue!/unsafe to-see))))))

(define (centralizer element group)
  (let ((operation (group-operation group)) (group (group-elements group)))
    (make-group (keep-matching-items
                 group
                 (lambda (x)
                   (equal? (operation element x) (operation x element))))
                operation)))

(define (conjugacy-class element the-group)
  (let ((operation (group-operation the-group))
        (group (group-elements the-group))
        (result (make-equal-hash-table))
        (inverses (make-equal-hash-table))
        (identity (get-identity the-group)))
    (let loop ((group group))
      (if (pair? group)
          (let ((inverse (hash-table/get inverses (car group) #f)))
            (if (not inverse)
                (begin
                  (set! inverse (group-inverse the-group (car group)))
                  (hash-table/put! inverses (car group) inverse)))
            (hash-table/put! result (operation (car group) (operation element inverse)) #t)
            (loop (cdr group)))
          (hash-table/key-list result)))))

(define (cosets subgroup group)
  (let ((operation (group-operation group))
        (group (group-elements group))
        (subgroup (group-elements subgroup))
        (table (make-equal-hash-table)))
    (let loop ((result '()) (group group) (coset-number 0))
      (if (pair? group)
          (let ((first-elt (operation (car group) (car subgroup))))
            (if (hash-table/get table first-elt #f)
                (loop result (cdr group) coset-number)
                (let ((coset (map (lambda (x) (operation (car group) x)) subgroup)))
                  (for-each
                   (lambda (x) (hash-table/put! table x coset-number))
                   coset)
                  (loop (cons coset result) (cdr group) (+ coset-number 1)))))
          (begin ; hash table needed for quotient group algorithm
            (set-tag! result 'coset-data table)
            (set-tag! result 'num-cosets coset-number)
            result)))))

(define (quotient group subgroup)
  (let* ((operation (group-operation group))
         (cosets (cosets subgroup group))
         (group (group-elements group))
         (subgroup (group-elements subgroup))
         (coset-hash (get-tag cosets 'coset-data #f))
         (num-cosets (get-tag cosets 'num-cosets #f))
         (result (make-equal-hash-table)))

    (let loop1 ((group group))
      (if (null? group)
          (make-group (hash-table/key-list result) apply-perm)
          (let ((perm (make-vector num-cosets #f)))
            (let loop2 ((cosets cosets))
              (if (pair? cosets)
                  (let ((coset (operation (car group) (caar cosets))))
                    (vector-set! perm
                                 (hash-table/get coset-hash (caar cosets) #f)
                                 (hash-table/get coset-hash coset #f))
                    (loop2 (cdr cosets)))
                  (begin
                    (hash-table/put! result (vector->list perm) '())
                    (loop1 (cdr group))))))))))

(define (powerset elts)
  (define (sometimes-prepend a l)
    (if (null? l)
        (list a '())
        (let next ((l l))
          (if (null? l)
              '()
              (cons (car l) (cons (cons a (car l)) (next (cdr l))))))))
  (let next ((elts elts))
    (if (null? elts)
        '(())
        (sometimes-prepend (car elts) (next (cdr elts))))))

(define (subgroups g)
  (filter (lambda (g)
            (and (not (null? (group-elements g)))
                 (group? g)))
          (map (lambda (set)
                 (make-group set (group-operation g)))
               (powerset (group-elements g)))))